# Fortuna: Random Value Generator
Fortuna's main goal is to provide a quick and easy way to build custom random-value generators.

The core functionality of Fortuna is based on the RNG Storm engine. While Storm is a high quality random engine, Fortuna is not appropriate for cryptography of any kind. Fortuna is meant for games, data science, A.I. and experimental programming, not security.

Suggested Installation: `$ pip install Fortuna`

Installation on platforms other than MacOS may require building from source files.


### Documentation Table of Contents:
- Numeric Limits
- Project Definitions
- Random Value Generators
- Random Integer Generators
- Random Index Generators
- Random Float Generators
- Random Bool Generator
- Shuffle Algorithms
- Test Suite Functions
- Test Suite Output
- Development Log
- Legal Information


##### Numeric Limits:
- Integer: 64 bit signed integer.
    - Input & Output Range: `(-2**63, 2**63)` or approximately +/- 9.2 billion billion.
    - Minimum: -9223372036854775807
    - Maximum:  9223372036854775807
- Float: 64 bit double precision floating point.
    - Minimum: -1.7976931348623157e+308
    - Maximum:  1.7976931348623157e+308
    - Epsilon Below Zero: -5e-324
    - Epsilon Above Zero:  5e-324

##### Project Definitions:
- Value: Any python object: str, int, list and lambda to name a few.
- Callable: Any callable object, function, method or lambda.
- Sequence: Any object that can be converted into a list via `list(some_object)`.
    - List, Tuple, Set, etc...
    - Comprehensions and Generators that produce Sequences also qualify.
    - Classes that wrap a simple collection will take any Sequence or Array.
    - Functions that operate on a simple collection will require an Array.
- Array: List, tuple or any object that inherits from either.
    - Must be an object that is indexed like a list.
    - List comprehensions qualify as Arrays, but sets and list generators do not.
    - All Arrays are Sequences but not all Sequences are Arrays.
- Pair: Array of two Values.
- Table: Array of Pairs.
    - List of lists of two values each.
    - Tuple of tuples of two values each.
    - Generators that produce Tables also qualify.
    - The result of zip(list_1, list_2) also qualifies.
- Matrix: Dictionary of Arrays.
    - Generators or comprehensions that produces a Matrix also qualify.
- Inclusive Range.
    - `[1, 10] -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Exclusive Range.
    - `(0, 11) -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Partial Ranges.
    - `[1, 11) -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
    - `(0, 10] -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Automatic Flattening. 
    - All random value generators in Fortuna will recursively call or "flatten" callable objects returned from the dataset at call time. 
    - Un-callable objects or those that require arguments will be returned in an uncalled state without error. 
    - A callable that can be flattened is any class initializer, function, method or lambda so long as it requires no arguments, it will be automatically flattened. 
    - Mixing callable objects with un-callable objects is fully supported, but it can be a bit messy. 
    - Nested callable objects are fully supported. Because `lambda(lambda) -> lambda` fixes everything for arbitrary values of 'because', 'fixes' and 'everything'. 
    - To disable automatic flattening, pass the optional keyword argument `flat=False`.


## Value Generators: Functional
`Fortuna.random_value(data: Array, flat=True) -> Value`
- @param data :: Array of Values.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Produces a random value from the list with a wide uniform distribution.

```python
from Fortuna import random_value


data = ("Apple", "Banana", "Cherry", "Grape", "Lime", "Orange")
print(random_value(data))  # prints a random value from the data, flat uniform distribution.
```

`Fortuna.cumulative_weighted_choice(weighted_table: Table, flat=True) -> Value`
- @param weighted_table :: Table of weighted Pairs. `[(1, "a"), (2, "b"), (3, "c")]`
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Produces a random value from the list with a custom distribution.

```python
from Fortuna import cumulative_weighted_choice


weighted_data = (
    (7, "Apple"),
    (11, "Banana"),
    (13, "Cherry"),
    (23, "Grape"),
    (26, "Lime"),
    (30, "Orange"),
)
print(cumulative_weighted_choice(weighted_data))
# prints a random value from the data, with probability based on cumulative weight.
```

`Fortuna.lazy_cat(data, range_to=0, func=random_index, flat=True) -> Value`
- The lazy_cat function is a more general form of the QuantumMonty class. Features dependency injection.
- @param data :: Array of Values.
- @param range_to :: Default is zero. Must be equal to or less than the length of data, this represents the size of the output distribution. When range_to == 0, the total length of data is used instead. This argument is passed to the input function to get a valid index into the data. When range_to is negative the back of the data will be considered the starting point.
- @param func :: Optional. Default is random_index. This callable must follow the ZeroCool method specifications. All built-in ZeroCool methods qualify.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Returns a random value from the data using the function you specify at call time.

`Fortuna.truffle_shuffle(data: list, flat=True) -> Value`
In-place micro-shuffle. Destructive. It is recommended to shuffle the list conventionally once before a series of truffle_shuffles.
- @param data :: list of Values. Must be a mutable list. For best results, each Value should be unique and the number of Values should be more than just a few.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Random value from the list with a wide uniform distribution. The average width of the output sequence will naturally scale up with the size of the list.

**Wide Uniform Distribution**: *"Wide"* refers to the average distance between consecutive occurrences of the same value in the output sequence. The goal of this type of distribution is to keep the output sequence free of clumps or streaks, while maintaining randomness and uniform probability. This is not the same as a *flat uniform distribution*. The two distributions should be statistically similar for any given set, but the output sequence repetitiveness will be very different.


## Value Generators: Classes
### TruffleShuffle
`Fortuna.TruffleShuffle(data: Sequence, flat=True) -> Callable -> Value`
- In-place micro-shuffle. Non-destructive, copies the list_of_values.
- @param data :: Sequence of Values. Automatically shuffles once with Knuth B at instantiation.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Callable Instance.
    - @return :: Random value from the list with a wide uniform distribution, performs an internal micro-shuffle. Pop a Value for return then use front_poisson insert to randomly put it back near the front of the list. The relative width of the output sequence will naturally scale up with the size of the list.


#### TruffleShuffle, Basic Use
```python
from Fortuna import TruffleShuffle


list_of_values = [1, 2, 3, 4, 5, 6]
truffle_shuffle = TruffleShuffle(list_of_values)

print(truffle_shuffle())
```

#### Automatic Flattening
```python
from Fortuna import TruffleShuffle


auto_flat = TruffleShuffle([lambda: 1, lambda: 2, lambda: 3])
print(auto_flat())  # will print the value 1, 2 or 3.
# Note: the lambda will not be called until call time and stays dynamic for the life of the object.

un_flat = TruffleShuffle([lambda: 1, lambda: 2, lambda: 3], flat=False)
print(un_flat()())  # will print the value 1, 2 or 3, mind the double-double parenthesis

auto_un_flat = TruffleShuffle([lambda x: x, lambda x: x + 1, lambda x:  x + 2], flat=False)
# Note: flat=False is not required here because the lambdas can not be called without input x satisfied.
# It is still recommended to specify flat=False if non-flat output is intend.
print(auto_un_flat()(1))  # will print the value 1, 2 or 3, mind the double-double parenthesis

```


#### Mixing Static Objects with Callable Objects
```python
from Fortuna import TruffleShuffle


""" With automatic flattening active, lambda() -> int can be treated as an int. """
mixed_flat = TruffleShuffle([1, 2, lambda: 3])  # this is fine and works as intended.
print(mixed_flat())  # will print 1, 2 or 3

mixed_un_flat = TruffleShuffle([1, 2, lambda: 3], flat=False) # this pattern is not recommended.
print(mixed_flat())  # will print 1, 2 or "Function <lambda at some_address>"
# This pattern is not recommended because you wont know the nature of what you get back.
# This is almost always not what you want, and it can give rise to messy logic in other areas of your code.
```


#### Dynamic Strings
To successfully express a dynamic string, and keep it dynamic, at least one level of indirection is required. Without an indirection the f-string would collapse into a static string too soon.

```python
from Fortuna import TruffleShuffle, d


# d() is a simple dice function, d(n) -> [1, n] flat uniform distribution.
dynamic_strings = TruffleShuffle((
    # while the probability of all A == all B == all C, individual probabilities of each possible string will differ based on the number of possible outputs of each category.
    lambda: f"A{d(2)}",  # -> A1 - A2, each are twice as likely as any particular B, and three times as likely as any C.
    lambda: f"B{d(4)}",  # -> B1 - B4, each are half as likely as any particular A, and 3/2 as likely as any C.
    lambda: f"C{d(6)}",  # -> C1 - C6, each are 1/3 as likely as any particular A and 2/3 as likely of any B.
))

print(dynamic_strings())  # prints a random dynamic string, flattened at call time.

"""
Sample Distribution of 10,000 dynamic_strings():
    A1: 16.92%
    A2: 16.66%
    B1: 8.08%
    B2: 8.51%
    B3: 8.15%
    B4: 8.1%
    C1: 5.62%
    C2: 5.84%
    C3: 5.71%
    C4: 5.43%
    C5: 5.22%
    C6: 5.76%
"""
```


### QuantumMonty
`Fortuna.QuantumMonty(data: Sequence, flat=True) -> Callable -> Value`
- @param data :: Sequence of Values.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Callable Object with Monty Methods for producing various distributions of the data.
    - @return :: Random value from the data. The instance will produce random values from the list using the selected distribution model or "monty". The default monty is the Quantum Monty Algorithm.

```python
from Fortuna import QuantumMonty


list_of_values = [1, 2, 3, 4, 5, 6]
monty = QuantumMonty(list_of_values)

print(monty())               # prints a random value from the list_of_values.
                             # uses the default Quantum Monty Algorithm.

print(monty.flat_uniform())  # prints a random value from the list_of_values.
                             # uses the "uniform" monty: a flat uniform distribution.
                             # equivalent to random.choice(list_of_values).
```
The QuantumMonty class represents a diverse collection of strategies for producing random values from a sequence where the output distribution is based on the method you choose. Generally speaking, each value in the sequence will have a probability that is based on its position in the sequence. For example: the "front" monty produces random values where the beginning of the sequence is geometrically more common than the back. Given enough samples the "front" monty will always converge to a 45 degree slope down for any list of unique values.

There are three primary method families: linear, gaussian, and poisson. Each family has three base methods; 'front', 'middle', 'back', plus a 'quantum' method that incorporates all three base methods. The quantum algorithms for each family produce distributions by overlapping the probability waves of the other methods in their family. The Quantum Monty Algorithm incorporates all nine base methods.

```python
import Fortuna


monty = Fortuna.QuantumMonty(
    ["Alpha", "Beta", "Delta", "Eta", "Gamma", "Kappa", "Zeta"]
)

# Each of the following methods will return a random value from the sequence.
# Each method has its own unique distribution model for the same data set.

""" Flat Base Case """
monty.flat_uniform()             # Flat Uniform Distribution

""" Geometric Positional """
monty.front_linear()        # Linear Descending, Triangle
monty.middle_linear()       # Linear Median Peak, Equilateral Triangle
monty.back_linear()         # Linear Ascending, Triangle
monty.quantum_linear()      # Linear Overlay, 3-way monty.

""" Gaussian Positional """
monty.front_gauss()         # Front Gamma
monty.middle_gauss()        # Scaled Gaussian
monty.back_gauss()          # Reversed Gamma
monty.quantum_gauss()       # Gaussian Overlay, 3-way monty.

""" Poisson Positional """
monty.front_poisson()       # 1/3 Mean Poisson
monty.middle_poisson()      # 1/2 Mean Poisson
monty.back_poisson()        # -1/3 Mean Poisson
monty.quantum_poisson()     # Poisson Overlay, 3-way monty.

""" Quantum Monty Algorithm """
monty()                     # Quantum Monty Algorithm, 9-way monty.
monty.quantum_monty()
```

### Weighted Choice: Custom Rarity
Weighted Choice offers two strategies for selecting random values from a sequence where programmable rarity is desired. Both produce a custom distribution of values based on the weights of the values.

The choice to use one strategy over the other is purely about which one suits you or your data best. Relative weights are easier to understand at a glance. However, many RPG Treasure Tables map rather nicely to a cumulative weighted strategy.

#### Cumulative Weight Strategy
`Fortuna.CumulativeWeightedChoice(weighted_table: Table, flat=True) -> Callable -> Value`
- @param weighted_table :: Table of weighted pairs.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Callable Instance
    - @return :: Random value from the weighted_table, distribution based on the weights of the values.

_Note: Logic dictates Cumulative Weights must be unique!_

```python
from Fortuna import CumulativeWeightedChoice


cum_weighted_choice = CumulativeWeightedChoice([
    (7, "Apple"),
    (11, "Banana"),
    (13, "Cherry"),
    (23, "Grape"),
    (26, "Lime"),
    (30, "Orange"),  # same as relative weight 4 because 30 - 26 = 4
])

print(cum_weighted_choice())  # prints a weighted random value
```

#### Relative Weight Strategy
`Fortuna.RelativeWeightedChoice(weighted_table: Table) -> Callable -> Value`
- @param weighted_table :: Table of weighted pairs.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Callable Instance
    - @return :: Random value from the weighted_table, distribution based on the weights of the values.

```python
from Fortuna import RelativeWeightedChoice


population = ["Apple", "Banana", "Cherry", "Grape", "Lime", "Orange"]
rel_weights = [7, 4, 2, 10, 3, 4]
rel_weighted_choice = RelativeWeightedChoice(zip(rel_weights, population))

print(rel_weighted_choice())  # prints a weighted random value
```

### FlexCat
`Fortuna.FlexCat(dict_of_lists: Matrix, key_bias="front_linear", val_bias="truffle_shuffle", flat=True) -> Callable -> Value`
- @param dict_of_lists :: Keyed Matrix of Value Sequences.
- @parm key_bias :: String indicating the name of the algorithm to use for random key selection.
- @parm val_bias :: String indicating the name of the algorithm to use for random value selection.
- @param flat :: Bool. Default: True. Option to flatten callable values. Automatic Flattening.
- @return :: Callable Instance
    - @param cat_key :: Optional. Default is None. Key selection by name. If specified, this will override the key_bias for a single call.
    - @return :: Value. Returns a random value generated with val_bias from a random sequence generated with key_bias.

FlexCat is like a two dimensional QuantumMonty.

The constructor takes two optional keyword arguments to specify the algorithms to be used to make random selections. The algorithm specified for selecting a key need not be the same as the one for selecting values. An optional key may be provided at call time to bypass the random key selection. Keys passed in this way must exactly match a key in the Matrix.

By default, FlexCat will use key_bias="front" and val_bias="truffle_shuffle", this will make the top of the data structure geometrically more common than the bottom and it will truffle shuffle the sequence values. This config is known as TopCat, it produces a descending-step, micro-shuffled distribution sequence. Many other combinations are available.

Algorithmic Options: _See QuantumMonty & TruffleShuffle for more details._
- "front_linear", Linear Descending
- "middle_linear", Linear Median Peak
- "back_linear", Linear Ascending
- "quantum_linear", Linear 3-way monty
- "front_gauss", Gamma Descending
- "middle_gauss", Scaled Gaussian
- "back_gauss", Gamma Ascending
- "quantum_gauss", Gaussian 3-way monty
- "front_poisson", Front 1/3 Mean Poisson
- "middle_poisson", Middle Mean Poisson
- "back_poisson", Back 1/3 Mean Poisson
- "quantum_poisson", Poisson 3-way monty
- "quantum_monty", Quantum Monty Algorithm, 9-way monty
- "flat_uniform", uniform flat distribution
- "truffle_shuffle", TruffleShuffle, wide uniform distribution

```python
from Fortuna import FlexCat


matrix_data = {
    "Cat_A": ("A1", "A2", "A3", "A4", "A5"),
    "Cat_B": ("B1", "B2", "B3", "B4", "B5"),
    "Cat_C": ("C1", "C2", "C3", "C4", "C5"),
}
flex_cat = FlexCat(matrix_data, key_bias="front_linear", val_bias="flat_uniform")

flex_cat()  # returns a "flat_uniform" random value from a random "front_linear" weighted category.
flex_cat("Cat_B")  # returns a "flat_uniform" random value specifically from the "Cat_B" sequence.
```

## Fortuna Functions
### Random Integer
`Fortuna.random_below(number: int) -> int`
- @param number :: Any Integer
- @return :: Returns a random integer in the range...
    - `random_below(number) -> [0, number)` for positive values.
    - `random_below(number) -> (number, 0]` for negative values.
    - `random_below(0) -> 0` Always returns zero when input is zero
- Flat uniform distribution.


`Fortuna.random_int(left_limit: int, right_limit: int) -> int`
- @param left_limit :: Any Integer
- @param right_limit :: Any Integer
- @return :: Returns a random integer in the range [left_limit, right_limit]
    - `random_int(1, 10) -> [1, 10]`
    - `random_int(10, 1) -> [1, 10]` same as above.
    - `random_int(A, B)` Always returns A when A == B
- Flat uniform distribution.


`Fortuna.random_range(start: int, stop: int = 0, step: int = 1) -> int`
- @param start :: Any Integer. Required starting point.
    - `random_range(10) -> [0, 10)` from 0 to 9. Same as `Fortuna.random_index(N)`
    - `random_range(-10) -> [-10, 0)` from -10 to -1. Same as `Fortuna.random_index(-N)`
- @param stop :: Any Integer. Optional stopping point. With at least two arguments, the order of the first two does not matter.
    - `random_range(0, 10) -> [0, 10)` from 0 to 9.
    - `random_range(10, 0) -> [0, 10)` same as above.
- @param step :: optional step size.
    - `random_range(0, 10, 2) -> [0, 10) by 2` even numbers from 0 to 8.
    - The sign of the step parameter controls the phase of the output. Negative stepping will flip the inclusively.
    - `random_range(0, 10, -1) -> (0, 10]` from 10 to 1.
    - `random_range(10, 0, -1) -> (0, 10]` same as above.
    - `random_range(10, 10, 0) -> [10]` a step size or range size of zero always returns the first parameter.
- @return :: Returns a random integer in the range [A, B) by increments of C.
- Flat uniform distribution.


`Fortuna.d(sides: int) -> int`
- Represents a single roll of a given size die.
- @param sides :: Any Integer. Represents the size or number of sides, most commonly six.
- @return :: Returns a random integer in the range [1, sides].
- Flat uniform distribution.


`Fortuna.dice(rolls: int, sides: int) -> int`
- Represents the sum of multiple rolls of the same size die.
- @param rolls :: Any Integer. Represents the number of times to roll the die.
- @param sides :: Any Integer. Represents the die size or number of sides, most commonly six.
- @return :: Returns a random integer in range [X, Y] where X = rolls and Y = rolls * sides.
- Geometric distribution based on the number and size of the dice rolled.
- Complexity scales primarily with the number of rolls, not the size of the dice.


`Fortuna.plus_or_minus(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Flat uniform distribution.


`Fortuna.plus_or_minus_linear(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Linear geometric, 45 degree triangle distribution centered on zero.


`Fortuna.plus_or_minus_gauss(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Stretched gaussian distribution centered on zero.


### Random Index, ZeroCool Specification
ZeroCool Methods are used by lazy_cat via dependency injection to generate random indices with any distribution.

ZeroCool methods must have the following properties:
Any distribution model is acceptable, so long as: The method or function takes exactly one parameter N such that: The method returns a random int in range `[0, N)` for positive values of N, and returns a random int in range `[N, 0)` for negative values of N. This symmetry matches how python will naturally index a list from the back for negative index values or from the front for positive index values.

ZeroCool functions often have an interesting limit as size goes to zero. ZeroCool does not place requirements the output of this input limit. At higher levels of abstraction, inside classes that employ ZeroCool methods- zero is always a sentinel to indicate the full range. In that case the length of the list is sent to the ZeroCool method, not zero. However for those who enjoy thinking a little deeper, consider the following:

If given the fact that an empty range is never an option, we could design a better solution than failure for input zero. Calculus might suggest that both infinity and negative infinity are equally viable output for an input limit of zero, but both are inappropriate for indexing a list. However, if we map infinity to the back of the list and minus infinity to the front of the list, then the following might hold: `random_index(0) -> [-1, 0]`. This "Top or Bottom" solution is not required for a method to be ZeroCool compatible, it is just an interesting option. Other valid possibilities include: always return None or 0 or -1 or throw an exception or spawn nasal demons, however none of these seem terribly helpful or useful. At least the Top/Bottom solution accurately reflects the "off by one" symmetry of the domain mapping that defines ZeroCool methods in general.


```python
from Fortuna import random_index


some_list = [i for i in range(100)]

print(some_list[random_index(10)])  # prints one of the first 10 items of some_list, [0, 9]
print(some_list[random_index(-10)])  # prints one of the last 10 items of some_list, [90, 99]
```
### ZeroCool Methods
`Fortuna.random_index(size: int) -> int` Flat uniform distribution

`Fortuna.front_gauss(size: int) -> int` Gamma Distribution: Front Peak

`Fortuna.middle_gauss(size: int) -> int` Stretched Gaussian Distribution: Median Peak

`Fortuna.back_gauss(size: int) -> int` Gamma Distribution: Back Peak

`Fortuna.quantum_gauss(size: int) -> int` Quantum Gaussian: Three-way Monty

`Fortuna.front_poisson(size: int) -> int` Poisson Distribution: Front 1/3 Peak

`Fortuna.middle_poisson(size: int) -> int` Poisson Distribution: Middle Peak

`Fortuna.back_poisson(size: int) -> int` Poisson Distribution: Back 1/3 Peak

`Fortuna.quantum_poisson(size: int) -> int` Quantum Poisson: Three-way Monty

`Fortuna.front_geometric(size: int) -> int` Linear Geometric: 45 Degree Front Peak

`Fortuna.middle_geometric(size: int) -> int` Linear Geometric: 45 Degree Middle Peak

`Fortuna.back_geometric(size: int) -> int` Linear Geometric: 45 Degree Back Peak

`Fortuna.quantum_geometric(size: int) -> int` Quantum Geometric: Three-way Monty

`Fortuna.quantum_monty(size: int) -> int` Quantum Monty: Twelve-way Monty


### Random Float Functions
`Fortuna.canonical() -> float`
- @return :: random float in range [0.0, 1.0), flat uniform.


`Fortuna.random_float(a: Float, b: Float) -> Float`
- @param a :: Float
- @param b :: Float
- @return :: random Float in range [a, b), flat uniform distribution.


### Random Truth
`Fortuna.percent_true(truth_factor: Float = 50.0) -> bool`
- @param truth_factor :: The probability of True as a percentage. Default is 50 percent.
- @return :: Produces True or False based on the truth_factor.
    - Always returns False if num is 0 or less
    - Always returns True if num is 100 or more.


### Conventional, In-place Shuffle Functions
`Fortuna.shuffle(array: list) -> None` 
- Knuth B shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.

`Fortuna.knuth(array: list) -> None` 
- Knuth A shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.

`Fortuna.fisher_yates(array: list) -> None` 
- Fisher-Yates shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.


### Test Suite Functions
`Fortuna.distribution_timer(func: staticmethod, *args, num_cycles=100000, **kwargs) -> None`

`Fortuna.quick_test(cycles=10000) -> None`


## Fortuna Distribution and Performance Test Suite
```
Fortuna Test Suite: RNG Storm Engine

Random Values:

Base Case
Output Distribution: Random.choice([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
Typical Timing: 952 nano seconds
Raw Samples: 1, 0, 9, 1, 7
Statistics of 10000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.5125
 Std Deviation: 2.8681817230851436
Distribution of 100000 Samples:
 0: 9.854%
 1: 9.944%
 2: 10.012%
 3: 10.031%
 4: 10.105%
 5: 10.076%
 6: 10.026%
 7: 9.834%
 8: 10.063%
 9: 10.055%

Output Distribution: random_value([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
Typical Timing: 156 nano seconds
Raw Samples: 3, 0, 4, 2, 3
Statistics of 10000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.5136
 Std Deviation: 2.8912715170143586
Distribution of 100000 Samples:
 0: 9.954%
 1: 10.044%
 2: 10.097%
 3: 10.04%
 4: 9.989%
 5: 10.111%
 6: 9.991%
 7: 9.867%
 8: 9.961%
 9: 9.946%

Output Analysis: TruffleShuffle(some_list)
Typical Timing: 375 nano seconds
Raw Samples: 0, 7, 6, 1, 2
Statistics of 10000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.463
 Std Deviation: 2.8696784672591016
Distribution of 100000 Samples:
 0: 10.016%
 1: 10.096%
 2: 10.028%
 3: 10.106%
 4: 10.046%
 5: 10.037%
 6: 9.967%
 7: 9.89%
 8: 9.888%
 9: 9.926%

Output Distribution: truffle_shuffle([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
Typical Timing: 218 nano seconds
Raw Samples: 9, 8, 4, 0, 5
Statistics of 10000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.5003
 Std Deviation: 2.859128778445835
Distribution of 100000 Samples:
 0: 10.045%
 1: 10.023%
 2: 9.903%
 3: 10.049%
 4: 10.034%
 5: 10.021%
 6: 9.944%
 7: 10.028%
 8: 9.962%
 9: 9.991%

Output Analysis: QuantumMonty(some_list)
Typical Timing: 343 nano seconds
Raw Samples: 7, 5, 7, 2, 4
Statistics of 10000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.5121
 Std Deviation: 2.780508355080284
Distribution of 100000 Samples:
 0: 10.358%
 1: 7.729%
 2: 8.573%
 3: 10.238%
 4: 13.062%
 5: 12.972%
 6: 10.226%
 7: 8.521%
 8: 7.953%
 9: 10.368%

Output Distribution: lazy_cat([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], zero_cool=<built-in function quantum_monty>)
Typical Timing: 312 nano seconds
Raw Samples: 8, 4, 7, 7, 2
Statistics of 10000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.4751
 Std Deviation: 2.7599169741444065
Distribution of 100000 Samples:
 0: 10.311%
 1: 8.049%
 2: 8.422%
 3: 10.258%
 4: 12.95%
 5: 12.945%
 6: 10.56%
 7: 8.5%
 8: 7.868%
 9: 10.137%

Base Case
Output Distribution: Random.choices([36, 30, 24, 18], cum_weights=[1, 10, 100, 1000])
Typical Timing: 1656 nano seconds
Raw Samples: [18], [18], [18], [18], [18]
Statistics of 10000 Samples:
 Minimum: 18
 Median: 18
 Maximum: 36
 Mean: 18.6576
 Std Deviation: 2.1042825476879408
Distribution of 100000 Samples:
 18: 90.007%
 24: 8.942%
 30: 0.946%
 36: 0.105%

Output Analysis: CumulativeWeightedChoice(zip(cum_weights, population))
Typical Timing: 250 nano seconds
Raw Samples: 18, 18, 18, 18, 18
Statistics of 10000 Samples:
 Minimum: 18
 Median: 18
 Maximum: 36
 Mean: 18.666
 Std Deviation: 2.1389954488387803
Distribution of 100000 Samples:
 18: 89.859%
 24: 9.077%
 30: 0.958%
 36: 0.106%

Base Case
Output Distribution: Random.choices([36, 30, 24, 18], weights=[1, 9, 90, 900])
Typical Timing: 2062 nano seconds
Raw Samples: [18], [18], [18], [18], [18]
Statistics of 10000 Samples:
 Minimum: 18
 Median: 18
 Maximum: 36
 Mean: 18.6666
 Std Deviation: 2.1159612872680165
Distribution of 100000 Samples:
 18: 90.107%
 24: 8.917%
 30: 0.861%
 36: 0.115%

Output Analysis: RelativeWeightedChoice(zip(rel_weights, population))
Typical Timing: 250 nano seconds
Raw Samples: 18, 18, 24, 24, 18
Statistics of 10000 Samples:
 Minimum: 18
 Median: 18
 Maximum: 36
 Mean: 18.6942
 Std Deviation: 2.1560499101647697
Distribution of 100000 Samples:
 18: 90.157%
 24: 8.799%
 30: 0.942%
 36: 0.102%

Output Analysis: FlexCat(some_matrix, key_bias="flat_uniform", val_bias="flat_uniform")
Typical Timing: 437 nano seconds
Raw Samples: 20, 300, 3, 200, 200
Statistics of 10000 Samples:
 Minimum: 1
 Median: 20
 Maximum: 400
 Mean: 91.6143
 Std Deviation: 129.21308191341834
Distribution of 100000 Samples:
 1: 8.41%
 2: 8.402%
 3: 8.18%
 4: 8.373%
 10: 8.439%
 20: 8.29%
 30: 8.385%
 40: 8.38%
 100: 8.302%
 200: 8.215%
 300: 8.315%
 400: 8.309%

Output Distribution: flex_cat({'A': [1, 2, 3, 4], 'B': [10, 20, 30, 40], 'C': [100, 200, 300, 400]})
Typical Timing: 781 nano seconds
Raw Samples: 30, 100, 400, 300, 200
Statistics of 10000 Samples:
 Minimum: 1
 Median: 20
 Maximum: 400
 Mean: 90.4296
 Std Deviation: 127.82586744895929
Distribution of 100000 Samples:
 1: 8.423%
 2: 8.254%
 3: 8.215%
 4: 8.361%
 10: 8.351%
 20: 8.42%
 30: 8.268%
 40: 8.267%
 100: 8.562%
 200: 8.233%
 300: 8.303%
 400: 8.343%


Random Integers:

Base Case
Output Distribution: Random.randrange(10)
Typical Timing: 843 nano seconds
Raw Samples: 3, 8, 5, 9, 0
Statistics of 10000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.4909
 Std Deviation: 2.848987339895755
Distribution of 100000 Samples:
 0: 9.992%
 1: 9.965%
 2: 10.142%
 3: 10.085%
 4: 10.067%
 5: 9.945%
 6: 9.994%
 7: 9.943%
 8: 10.006%
 9: 9.861%

Output Distribution: random_below(10)
Typical Timing: 62 nano seconds
Raw Samples: 3, 9, 9, 1, 3
Statistics of 10000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.4999
 Std Deviation: 2.8639867723198726
Distribution of 100000 Samples:
 0: 9.976%
 1: 10.089%
 2: 10.133%
 3: 10.09%
 4: 9.861%
 5: 9.937%
 6: 9.994%
 7: 10.026%
 8: 9.946%
 9: 9.948%

Output Distribution: random_index(10)
Typical Timing: 62 nano seconds
Raw Samples: 9, 9, 8, 7, 3
Statistics of 10000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.5394
 Std Deviation: 2.8736863863937443
Distribution of 100000 Samples:
 0: 10.082%
 1: 9.828%
 2: 10.061%
 3: 10.002%
 4: 9.938%
 5: 9.811%
 6: 10.278%
 7: 10.074%
 8: 10.026%
 9: 9.9%

Output Distribution: random_range(10)
Typical Timing: 93 nano seconds
Raw Samples: 1, 7, 9, 0, 8
Statistics of 10000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.4448
 Std Deviation: 2.873495894875402
Distribution of 100000 Samples:
 0: 10.102%
 1: 10.09%
 2: 9.995%
 3: 10.046%
 4: 9.862%
 5: 9.955%
 6: 10.041%
 7: 10.14%
 8: 9.888%
 9: 9.881%

Output Distribution: random_below(-10)
Typical Timing: 93 nano seconds
Raw Samples: -7, -1, 0, -5, -3
Statistics of 10000 Samples:
 Minimum: -9
 Median: -5
 Maximum: 0
 Mean: -4.4903
 Std Deviation: 2.8713290238209037
Distribution of 100000 Samples:
 -9: 10.122%
 -8: 10.035%
 -7: 10.044%
 -6: 9.88%
 -5: 9.999%
 -4: 9.852%
 -3: 10.135%
 -2: 9.965%
 -1: 9.991%
 0: 9.977%

Output Distribution: random_index(-10)
Typical Timing: 77 nano seconds
Raw Samples: -6, -1, -5, -2, -6
Statistics of 10000 Samples:
 Minimum: -8
 Median: -4
 Maximum: 1
 Mean: -3.5104
 Std Deviation: 2.8812014806669146
Distribution of 100000 Samples:
 -8: 9.893%
 -7: 10.01%
 -6: 10.06%
 -5: 10.025%
 -4: 10.083%
 -3: 9.94%
 -2: 9.804%
 -1: 10.092%
 0: 9.928%
 1: 10.165%

Output Distribution: random_range(-10)
Typical Timing: 93 nano seconds
Raw Samples: -1, -9, -6, -2, -2
Statistics of 10000 Samples:
 Minimum: -10
 Median: -6
 Maximum: -1
 Mean: -5.5504
 Std Deviation: 2.8874025610148193
Distribution of 100000 Samples:
 -10: 10.112%
 -9: 10.025%
 -8: 9.854%
 -7: 9.953%
 -6: 9.981%
 -5: 10.055%
 -4: 9.997%
 -3: 10.03%
 -2: 9.844%
 -1: 10.149%

Base Case
Output Distribution: Random.randrange(1, 10)
Typical Timing: 1062 nano seconds
Raw Samples: 8, 2, 2, 8, 6
Statistics of 10000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 4.9957
 Std Deviation: 2.5802223285458314
Distribution of 100000 Samples:
 1: 11.015%
 2: 11.066%
 3: 11.13%
 4: 11.216%
 5: 11.048%
 6: 11.123%
 7: 11.218%
 8: 11.129%
 9: 11.055%

Output Distribution: random_range(1, 10)
Typical Timing: 93 nano seconds
Raw Samples: 9, 6, 8, 5, 8
Statistics of 10000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 5.0102
 Std Deviation: 2.59583701279177
Distribution of 100000 Samples:
 1: 10.837%
 2: 11.232%
 3: 11.083%
 4: 11.039%
 5: 11.096%
 6: 11.211%
 7: 11.299%
 8: 11.123%
 9: 11.08%

Output Distribution: random_range(10, 1)
Typical Timing: 93 nano seconds
Raw Samples: 4, 6, 3, 8, 5
Statistics of 10000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 5.0212
 Std Deviation: 2.5656985093506384
Distribution of 100000 Samples:
 1: 10.992%
 2: 11.174%
 3: 10.959%
 4: 11.464%
 5: 11.139%
 6: 11.067%
 7: 11.018%
 8: 10.972%
 9: 11.215%

Base Case
Output Distribution: Random.randint(-5, 5)
Typical Timing: 1156 nano seconds
Raw Samples: -1, -5, 1, 3, 2
Statistics of 10000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: 0.0484
 Std Deviation: 3.1888358859356094
Distribution of 100000 Samples:
 -5: 9.065%
 -4: 9.089%
 -3: 9.082%
 -2: 9.202%
 -1: 9.072%
 0: 9.073%
 1: 9.062%
 2: 9.066%
 3: 9.058%
 4: 9.065%
 5: 9.166%

Output Distribution: random_int(-5, 5)
Typical Timing: 62 nano seconds
Raw Samples: 0, 0, 0, -1, 3
Statistics of 10000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: 0.0671
 Std Deviation: 3.163731737971945
Distribution of 100000 Samples:
 -5: 9.22%
 -4: 9.062%
 -3: 9.14%
 -2: 9.112%
 -1: 9.102%
 0: 8.968%
 1: 8.985%
 2: 9.072%
 3: 8.984%
 4: 9.199%
 5: 9.156%

Base Case
Output Distribution: Random.randrange(1, 20, 2)
Typical Timing: 1250 nano seconds
Raw Samples: 7, 17, 15, 3, 11
Statistics of 10000 Samples:
 Minimum: 1
 Median: 9
 Maximum: 19
 Mean: 9.892
 Std Deviation: 5.699226624587922
Distribution of 100000 Samples:
 1: 9.805%
 3: 10.012%
 5: 10.186%
 7: 10.125%
 9: 9.958%
 11: 10.032%
 13: 10.023%
 15: 9.988%
 17: 9.881%
 19: 9.99%

Output Distribution: random_range(1, 20, 2)
Typical Timing: 93 nano seconds
Raw Samples: 15, 9, 19, 17, 3
Statistics of 10000 Samples:
 Minimum: 1
 Median: 11
 Maximum: 19
 Mean: 10.0936
 Std Deviation: 5.735375183268247
Distribution of 100000 Samples:
 1: 10.057%
 3: 10.065%
 5: 10.049%
 7: 9.892%
 9: 9.836%
 11: 10.18%
 13: 10.056%
 15: 9.957%
 17: 9.857%
 19: 10.051%

Output Distribution: random_range(1, 20, -2)
Typical Timing: 93 nano seconds
Raw Samples: 16, 14, 4, 6, 14
Statistics of 10000 Samples:
 Minimum: 2
 Median: 10
 Maximum: 20
 Mean: 10.9678
 Std Deviation: 5.722013395976138
Distribution of 100000 Samples:
 2: 10.066%
 4: 10.014%
 6: 9.982%
 8: 10.051%
 10: 10.093%
 12: 9.967%
 14: 9.944%
 16: 9.899%
 18: 10.192%
 20: 9.792%

Output Distribution: d(10)
Typical Timing: 62 nano seconds
Raw Samples: 9, 2, 8, 7, 10
Statistics of 10000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 10
 Mean: 5.509
 Std Deviation: 2.878288980409277
Distribution of 100000 Samples:
 1: 10.118%
 2: 9.968%
 3: 9.901%
 4: 10.044%
 5: 9.922%
 6: 9.992%
 7: 9.978%
 8: 10.121%
 9: 9.963%
 10: 9.993%

Output Distribution: dice(3, 6)
Typical Timing: 93 nano seconds
Raw Samples: 12, 9, 15, 7, 10
Statistics of 10000 Samples:
 Minimum: 3
 Median: 11
 Maximum: 18
 Mean: 10.535
 Std Deviation: 2.9712047734514324
Distribution of 100000 Samples:
 3: 0.48%
 4: 1.396%
 5: 2.857%
 6: 4.608%
 7: 7.034%
 8: 9.772%
 9: 11.617%
 10: 12.576%
 11: 12.45%
 12: 11.753%
 13: 9.528%
 14: 6.838%
 15: 4.606%
 16: 2.675%
 17: 1.368%
 18: 0.442%

Output Distribution: ability_dice(4)
Typical Timing: 218 nano seconds
Raw Samples: 13, 8, 12, 12, 14
Statistics of 10000 Samples:
 Minimum: 3
 Median: 12
 Maximum: 18
 Mean: 12.2913
 Std Deviation: 2.8719800028749662
Distribution of 100000 Samples:
 3: 0.079%
 4: 0.327%
 5: 0.777%
 6: 1.706%
 7: 2.977%
 8: 4.797%
 9: 7.034%
 10: 9.429%
 11: 11.556%
 12: 12.848%
 13: 13.187%
 14: 12.272%
 15: 10.04%
 16: 7.229%
 17: 4.131%
 18: 1.611%

Output Distribution: plus_or_minus(5)
Typical Timing: 93 nano seconds
Raw Samples: -5, -2, -4, 1, -3
Statistics of 10000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: 0.0042
 Std Deviation: 3.174963054751961
Distribution of 100000 Samples:
 -5: 9.184%
 -4: 9.294%
 -3: 9.112%
 -2: 9.045%
 -1: 8.957%
 0: 8.978%
 1: 9.317%
 2: 8.992%
 3: 9.02%
 4: 9.1%
 5: 9.001%

Output Distribution: plus_or_minus_linear(5)
Typical Timing: 93 nano seconds
Raw Samples: 2, -3, -3, 1, -3
Statistics of 10000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.0005
 Std Deviation: 2.408543098635729
Distribution of 100000 Samples:
 -5: 2.773%
 -4: 5.468%
 -3: 8.226%
 -2: 11.125%
 -1: 13.83%
 0: 16.716%
 1: 13.985%
 2: 11.276%
 3: 8.4%
 4: 5.388%
 5: 2.813%

Output Distribution: plus_or_minus_gauss(5)
Typical Timing: 93 nano seconds
Raw Samples: -2, -1, -1, 0, -2
Statistics of 10000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.0082
 Std Deviation: 1.2325926693230196
Distribution of 100000 Samples:
 -5: 0.083%
 -4: 0.493%
 -3: 2.435%
 -2: 7.337%
 -1: 15.986%
 0: 47.366%
 1: 16.037%
 2: 7.343%
 3: 2.33%
 4: 0.526%
 5: 0.064%


Random Floats:

Output Distribution: canonical()
Typical Timing: 62 nano seconds
Raw Samples: 0.7745709793090583, 0.9804306624612935, 0.13011799075080774, 0.7008239345083678, 0.5978009828511501
Statistics of 10000 Samples:
 Minimum: 7.237864837890157e-06
 Median: (0.5007331171202822, 0.5010455525738319)
 Maximum: 0.9998298367334093
 Mean: 0.5025309507333375
 Std Deviation: 0.2890272203030065
Post-processor Distribution of 100000 Samples using round method:
 0: 50.088%
 1: 49.912%

Output Distribution: random_float(0.0, 10.0)
Typical Timing: 62 nano seconds
Raw Samples: 1.997210953155215, 8.471452268646235, 5.405259143351109, 8.324680130788405, 4.365916194133978
Statistics of 10000 Samples:
 Minimum: 0.0006866180274310552
 Median: (5.0873279907765205, 5.087455522837279)
 Maximum: 9.998943831469749
 Mean: 5.038347348684213
 Std Deviation: 2.8735004010208773
Post-processor Distribution of 100000 Samples using floor method:
 0: 9.893%
 1: 9.99%
 2: 9.856%
 3: 10.039%
 4: 9.992%
 5: 10.119%
 6: 10.066%
 7: 10.083%
 8: 9.949%
 9: 10.013%


Random Booleans:

Output Distribution: percent_true(33.33)
Typical Timing: 62 nano seconds
Raw Samples: False, True, False, False, True
Statistics of 10000 Samples:
 Minimum: False
 Median: False
 Maximum: True
 Mean: 0.3375
 Std Deviation: 0.4728806526082042
Distribution of 100000 Samples:
 False: 66.596%
 True: 33.404%


Shuffle Performance Tests:

Base Case: random.shuffle(some_list_100)
Typical Timing: 67000 nano seconds

fisher_yates(some_list_100)
Typical Timing: 6906 nano seconds

knuth(some_list_100)
Typical Timing: 6968 nano seconds

shuffle(some_list_100)
Typical Timing: 3968 nano seconds

-------------------------------------------------------------------------
Total Test Time: 3.122 seconds

```


## Fortuna Development Log
##### Fortuna 3.3.7
- Fixed Performance Bug: ZeroCool Linear Family, affects QuantumMonty and FlexCat

##### Fortuna 3.3.6
- Docs Update

##### Fortuna 3.3.5
- ABI Updates
- Bug Fixes

##### Fortuna 3.3.4
- Examples Update

##### Fortuna 3.3.3
- Test Suite Update

##### Fortuna 3.3.2 - internal
- Documentation Update

##### Fortuna 3.3.1 - internal
- Minor Bug Fix

##### Fortuna 3.3.0 - internal
- Added `plus_or_minus_gauss(N: int) -> int` random int in range [-N, N] Stretched Gaussian Distribution

##### Fortuna 3.2.3
- Small Typos Fixed

##### Fortuna 3.2.2
- Documentation update.

##### Fortuna 3.2.1
- Small Typo Fixed

##### Fortuna 3.2.0
- API updates:
    - QunatumMonty.uniform -> QunatumMonty.flat_uniform
    - QunatumMonty.front -> QunatumMonty.front_linear
    - QunatumMonty.middle -> QunatumMonty.middle_linear
    - QunatumMonty.back -> QunatumMonty.back_linear
    - QunatumMonty.quantum -> QunatumMonty.quantum_linear
    - randindex -> random_index
    - randbelow -> random_below
    - randrange -> random_range
    - randint   -> random_int

##### Fortuna 3.1.0
- `discrete()` has been removed, see Weighted Choice.
- `lazy_cat()` added.
- All ZeroCool methods have been raised to top level API, for use with lazy_cat()

##### Fortuna 3.0.1
- minor typos.

##### Fortuna 3.0.0
- Storm 2 Rebuild.

##### Fortuna 2.1.1
- Small bug fixes.
- Test updates.

##### Fortuna 2.1.0, Major Feature Update
- Fortuna now includes the best of RNG and Pyewacket.

##### Fortuna 2.0.3
- Bug fix.

##### Fortuna 2.0.2
- Clarified some documentation.

##### Fortuna 2.0.1
- Fixed some typos.

##### Fortuna 2.0.0b1-10
- Total rebuild. New RNG Storm Engine.

##### Fortuna 1.26.7.1
- README updated.

##### Fortuna 1.26.7
- Small bug fix.

##### Fortuna 1.26.6
- Updated README to reflect recent changes to the test script.

##### Fortuna 1.26.5
- Fixed small bug in test script.

##### Fortuna 1.26.4
- Updated documentation for clarity.
- Fixed a minor typo in the test script.

##### Fortuna 1.26.3
- Clean build.

##### Fortuna 1.26.2
- Fixed some minor typos.

##### Fortuna 1.26.1
- Release.

##### Fortuna 1.26.0 beta 2
- Moved README and LICENSE files into fortuna_extras folder.

##### Fortuna 1.26.0 beta 1
- Dynamic version scheme implemented.
- The Fortuna Extension now requires the fortuna_extras package, previously it was optional.

##### Fortuna 1.25.4
- Fixed some minor typos in the test script.

##### Fortuna 1.25.3
- Since version 1.24 Fortuna requires Python 3.7 or higher. This patch corrects an issue where the setup script incorrectly reported requiring Python 3.6 or higher.

##### Fortuna 1.25.2
- Updated test suite.
- Major performance update for TruffleShuffle.
- Minor performance update for QuantumMonty & FlexCat: cycle monty.

##### Fortuna 1.25.1
- Important bug fix for TruffleShuffle, QuantumMonty and FlexCat.

##### Fortuna 1.25
- Full 64bit support.
- The Distribution & Performance Tests have been redesigned.
- Bloat Control: Two experimental features have been removed.
    - RandomWalk
    - CatWalk
- Bloat Control: Several utility functions have been removed from the top level API. These function remain in the Fortuna namespace for now, but may change in the future without warning.
    - stretch_bell, internal only.
    - min_max, not used anymore.
    - analytic_continuation, internal only.
    - flatten, internal only.

##### Fortuna 1.24.3
- Low level refactoring, non-breaking patch.

##### Fortuna 1.24.2
- Setup config updated to improve installation.

##### Fortuna 1.24.1
- Low level patch to avoid potential ADL issue. All low level function calls are now qualified.

##### Fortuna 1.24
- Documentation updated for even more clarity.
- Bloat Control: Two naïve utility functions that are no longer used in the module have been removed.
    - n_samples -> use a list comprehension instead. `[f(x) for _ in range(n)]`
    - bind -> use a lambda instead. `lambda: f(x)`

##### Fortuna 1.23.7
- Documentation updated for clarity.
- Minor bug fixes.
- TruffleShuffle has been redesigned slightly, it now uses a random rotate instead of swap.
- Custom `__repr__` methods have been added to each class.

##### Fortuna 1.23.6
- New method for QuantumMonty: quantum_not_monty - produces the upside down quantum_monty.
- New bias option for FlexCat: not_monty.

##### Fortuna 1.23.5.1
- Fixed some small typos.

##### Fortuna 1.23.5
- Documentation updated for clarity.
- All sequence wrappers can now accept generators as input.
- Six new functions added:
    - random_float() -> float in range [0.0..1.0) exclusive, uniform flat distribution.
    - percent_true_float(num: float) -> bool, Like percent_true but with floating point precision.
    - plus_or_minus_linear_down(num: int) -> int in range [-num..num], upside down pyramid.
    - plus_or_minus_curve_down(num: int) -> int in range [-num..num], upside down bell curve.
    - mostly_not_middle(num: int) -> int in range [0..num], upside down pyramid.
    - mostly_not_center(num: int) -> int in range [0..num], upside down bell curve.
- Two new methods for QuantumMonty:
    - mostly_not_middle
    - mostly_not_center
- Two new bias options for FlexCat, either can be used to define x and/or y axis bias:
    - not_middle
    - not_center

##### Fortuna 1.23.4.2
- Fixed some minor typos in the README.md file.

##### Fortuna 1.23.4.1
- Fixed some minor typos in the test suite.

##### Fortuna 1.23.4
- Fortuna is now Production/Stable!
- Fortuna and Fortuna Pure now use the same test suite.

##### Fortuna 0.23.4, first release candidate.
- RandomCycle, BlockCycle and TruffleShuffle have been refactored and combined into one class: TruffleShuffle.
- QuantumMonty and FlexCat will now use the new TruffleShuffle for cycling.
- Minor refactoring across the module.

##### Fortuna 0.23.3, internal
- Function shuffle(arr: list) added.

##### Fortuna 0.23.2, internal
- Simplified the plus_or_minus_curve(num: int) function, output will now always be bounded to the range [-num..num].
- Function stretched_bell(num: int) added, this matches the previous behavior of an unbounded plus_or_minus_curve.

##### Fortuna 0.23.1, internal
- Small bug fixes and general clean up.

##### Fortuna 0.23.0
- The number of test cycles in the test suite has been reduced to 10,000 (down from 100,000). The performance of the pure python implementation and the c-extension are now directly comparable.
- Minor tweaks made to the examples in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.22.2, experimental features
- BlockCycle class added.
- RandomWalk class added.
- CatWalk class added.

##### Fortuna 0.22.1
- Fortuna classes no longer return lists of values, this behavior has been extracted to a free function called n_samples.

##### Fortuna 0.22.0, experimental features
- Function bind added.
- Function n_samples added.

##### Fortuna 0.21.3
- Flatten will no longer raise an error if passed a callable item that it can't call. It correctly returns such items in an uncalled state without error.
- Simplified `.../fortuna_extras/fortuna_examples.py` - removed unnecessary class structure.

##### Fortuna 0.21.2
- Fix some minor bugs.

##### Fortuna 0.21.1
- Fixed a bug in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.21.0
- Function flatten added.
- Flatten: The Fortuna classes will recursively unpack callable objects in the data set.

##### Fortuna 0.20.10
- Documentation updated.

##### Fortuna 0.20.9
- Minor bug fixes.

##### Fortuna 0.20.8, internal
- Testing cycle for potential new features.

##### Fortuna 0.20.7
- Documentation updated for clarity.

##### Fortuna 0.20.6
- Tests updated based on recent changes.

##### Fortuna 0.20.5, internal
- Documentation updated based on recent changes.

##### Fortuna 0.20.4, internal
- WeightedChoice (both types) can optionally return a list of samples rather than just one value, control the length of the list via the n_samples argument.

##### Fortuna 0.20.3, internal
- RandomCycle can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.2, internal
- QuantumMonty can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.1, internal
- FlexCat can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.0, internal
- FlexCat now accepts a standard dict as input. The ordered(ness) of dict is now part of the standard in Python 3.7.1. Previously FlexCat required an OrderedDict, now it accepts either and treats them the same.

##### Fortuna 0.19.7
- Fixed bug in `.../fortuna_extras/fortuna_examples.py`.

##### Fortuna 0.19.6
- Updated documentation formatting.
- Small performance tweak for QuantumMonty and FlexCat.

##### Fortuna 0.19.5
- Minor documentation update.

##### Fortuna 0.19.4
- Minor update to all classes for better debugging.

##### Fortuna 0.19.3
- Updated plus_or_minus_curve to allow unbounded output.

##### Fortuna 0.19.2
- Internal development cycle.
- Minor update to FlexCat for better debugging.

##### Fortuna 0.19.1
- Internal development cycle.

##### Fortuna 0.19.0
- Updated documentation for clarity.
- MultiCat has been removed, it is replaced by FlexCat.
- Mostly has been removed, it is replaced by QuantumMonty.

##### Fortuna 0.18.7
- Fixed some more README typos.

##### Fortuna 0.18.6
- Fixed some README typos.

##### Fortuna 0.18.5
- Updated documentation.
- Fixed another minor test bug.

##### Fortuna 0.18.4
- Updated documentation to reflect recent changes.
- Fixed some small test bugs.
- Reduced default number of test cycles to 10,000 - down from 100,000.

##### Fortuna 0.18.3
- Fixed some minor README typos.

##### Fortuna 0.18.2
- Fixed a bug with Fortuna Pure.

##### Fortuna 0.18.1
- Fixed some minor typos.
- Added tests for `.../fortuna_extras/fortuna_pure.py`

##### Fortuna 0.18.0
- Introduced new test format, now includes average call time in nanoseconds.
- Reduced default number of test cycles to 100,000 - down from 1,000,000.
- Added pure Python implementation of Fortuna: `.../fortuna_extras/fortuna_pure.py`
- Promoted several low level functions to top level.
    - `zero_flat(num: int) -> int`
    - `zero_cool(num: int) -> int`
    - `zero_extreme(num: int) -> int`
    - `max_cool(num: int) -> int`
    - `max_extreme(num: int) -> int`
    - `analytic_continuation(func: staticmethod, num: int) -> int`
    - `min_max(num: int, lo: int, hi: int) -> int`

##### Fortuna 0.17.3
- Internal development cycle.

##### Fortuna 0.17.2
- User Requested: dice() and d() functions now support negative numbers as input.

##### Fortuna 0.17.1
- Fixed some minor typos.

##### Fortuna 0.17.0
- Added QuantumMonty to replace Mostly, same default behavior with more options.
- Mostly is depreciated and may be removed in a future release.
- Added FlexCat to replace MultiCat, same default behavior with more options.
- MultiCat is depreciated and may be removed in a future release.
- Expanded the Treasure Table example in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.16.2
- Minor refactoring for WeightedChoice.

##### Fortuna 0.16.1
- Redesigned fortuna_examples.py to feature a dynamic random magic item generator.
- Raised cumulative_weighted_choice function to top level.
- Added test for cumulative_weighted_choice as free function.
- Updated MultiCat documentation for clarity.

##### Fortuna 0.16.0
- Pushed distribution_timer to the .pyx layer.
- Changed default number of iterations of tests to 1 million, up form 1 hundred thousand.
- Reordered tests to better match documentation.
- Added Base Case Fortuna.fast_rand_below.
- Added Base Case Fortuna.fast_d.
- Added Base Case Fortuna.fast_dice.

##### Fortuna 0.15.10
- Internal Development Cycle

##### Fortuna 0.15.9
- Added Base Cases for random_value()
- Added Base Case for randint()

##### Fortuna 0.15.8
- Clarified MultiCat Test

##### Fortuna 0.15.7
- Fixed minor typos.

##### Fortuna 0.15.6
- Fixed minor typos.
- Simplified MultiCat example.

##### Fortuna 0.15.5
- Added MultiCat test.
- Fixed some minor typos in docs.

##### Fortuna 0.15.4
- Performance optimization for both WeightedChoice() variants.
- Cython update provides small performance enhancement across the board.
- Compilation now leverages Python3 all the way down.
- MultiCat pushed to the .pyx layer for better performance.

##### Fortuna 0.15.3
- Reworked the MultiCat example to include several randomizing strategies working in concert.
- Added Multi Dice 10d10 performance tests.
- Updated sudo code in documentation to be more pythonic.

##### Fortuna 0.15.2
- Fixed: Linux installation failure.
- Added: complete source files to the distribution (.cpp .hpp .pyx).

##### Fortuna 0.15.1
- Updated & simplified distribution_timer in `fortuna_tests.py`
- Readme updated, fixed some typos.
- Known issue preventing successful installation on some linux platforms.

##### Fortuna 0.15.0
- Performance tweaks.
- Readme updated, added some details.

##### Fortuna 0.14.1
- Readme updated, fixed some typos.

##### Fortuna 0.14.0
- Fixed a bug where the analytic continuation algorithm caused a rare issue during compilation on some platforms.

##### Fortuna 0.13.3
- Fixed Test Bug: percent sign was missing in output distributions.
- Readme updated: added update history, fixed some typos.

##### Fortuna 0.13.2
- Readme updated for even more clarity.

##### Fortuna 0.13.1
- Readme updated for clarity.

##### Fortuna 0.13.0
- Minor Bug Fixes.
- Readme updated for aesthetics.
- Added Tests: `.../fortuna_extras/fortuna_tests.py`

##### Fortuna 0.12.0
- Internal test for future update.

##### Fortuna 0.11.0
- Initial Release: Public Beta

##### Fortuna 0.10.0
- Module name changed from Dice to Fortuna

##### Dice 0.1.x - 0.9.x
- Experimental Phase


## Legal Information
Fortuna © 2019 Broken aka Robert W Sharp, all rights reserved.

Fortuna is licensed under a Creative Commons Attribution-NonCommercial 3.0 Unported License.

See online version of this license here: <http://creativecommons.org/licenses/by-nc/3.0/>
